#!/bin/bash

#     Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
#     The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# 
#    Copyright (c) 2012-2013 by Deon George

# Check our arguments
while getopts ":d:DF:m:M:p:s:u:" opt; do
	case $opt in
		# Just backup this DB
		d) DB=$OPTARG ;;

		D) DEBUG=1 ;;

		# Filespace to store in TSM
		F) FILESPACE=$OPTARG ;;

		# Digest to use
		M) DIGEST=$OPTARG ;;

		# Optional MYSQL arguments
		m) MYSQL_ARGS=$OPTARG ;;

		# Mysql PASSWORD
		p) DBPASS=$OPTARG ;;

		# Skip DBs
		s) SKIP_DB=$OPTARG ;;

		# Mysql USERNAME
		u) DBUSER=$OPTARG ;;

		# Error Checking
		\?)
			echo "Invalid option: -$OPTARG" >&2
			exit 1
			;;
		:)
			echo "Option -$OPTARG requires an argument." >&2
			exit 1
			;;
	esac
done

# Our defaults
FILESPACE=${FILESPACE:-"/mysqldump"}
MYSQL_ARGS=${MYSQL_ARGS:-"-FlK --skip-extended-insert --skip-dump-date --log-error=/var/log/tsm/tdp.mysql.log"}

[ -n "$DBPASS" ] && MYSQL_AUTH="-p${DBPASS}"
[ -n "$DBUSER" ] && MYSQL_AUTH="-u${DBUSER} ${MYSQL_AUTH}"

# List our Databases
function listdb() {
	echo "show databases" |mysql -N ${MYSQL_AUTH}
}

# Get Size
function getsize() {
	echo "SELECT sum(data_length+index_length) FROM information_schema.TABLES where table_schema=\"$1\" GROUP BY table_schema ; " | mysql -N ${MYSQL_AUTH}
}

# TSMPIPE ARGUMENTS
[ -n "$DIGEST" ] && TSMPIPE_ARGS="${TSMPIPE_ARGS} -m ${DIGEST}"

listdb | while read db; do
	[ -n "$DB" -a "$db" != "$DB" ] && continue;

	SKIP=0
	if [ -n "$SKIP_DB" ]; then
		OLDIFS=$IFS
		IFS=":"
		for x in $SKIP_DB ; do
			[ $x == $db ] && SKIP=1 && break
		done
		IFS=$OLDIFS
	fi

	# We need to skip this DB
	[ "$SKIP" -eq 1 ] && continue;

	size=$(getsize $db);

	# We need to skip this DB if it is empty
	if [ -z "$size" ]; then
		[ -n "$DEBUG" ] && echo "Skipping DB ($db), with no size ($size)"
		continue
	fi

	[ -n "$DEBUG" ] && echo "Processing DB ($db), with size ($size)"

	# Backup to TSM
	mysqldump ${MYSQL_AUTH} ${MYSQL_ARGS} $db| tsmpipe -Bcs ${FILESPACE}/${db} -f FULL -L ${size} ${TSMPIPE_ARGS}
done;
